import static java.lang.Math.PI;
import static java.lang.Math.cos;
import static java.lang.Math.sin;
import static java.lang.Math.tan;
		
import java.util.function.*;

public class FindRootsWithLambda {
    /* don't need to define a method to represent the function
	 * roots of which will be searched */

    /**
     * Find the root of the function f(x) with the range [a,b] using
     * a simple bisection method.
     * @param f function as an instance of Function<T,R> interface
     * @param a minimum value of x to test
     * @param b maximum value of x to test
     * @param e tolerance - allowed size of interval within which root is found
     * @return x such that f(x+epsilon) == 0 where epsilon < e
     */
    public static double bisect(Function<Double, Double> f, double a, 
				double b, double e) {
        if (Math.signum(f.apply(a)) == Math.signum(f.apply(b))) throw new 
			IllegalArgumentException("sign is same at both endpoints");
        
        while(Math.abs(b-a) >= e) {
           double m = (a+b) / 2.0;
           if (Math.signum(f.apply(m)) == Math.signum(f.apply(a))) {
               a = m;
           } else {
               b = m;
           }
        }
        return a;
    }

    public static void main(String[] args) {
        double e = 1.0e-15;
		
        // A simple example function f(x) = x^2 - x - 6.
        // Because f(x) = (x+2)(x-3), there are roots at -2.0 and 3.0
        double root = bisect(x -> x*x - x - 6, 0.0, 4.0, e);
        System.out.println(root); // should print something close to 3.0
		root = bisect(x -> cos(x) - 0.5, 0, PI/2.0, e);
		System.out.println(root); 
		root = bisect(x -> tan(x) - 1, PI/4, PI/2.0 - e, e);
		System.out.printf("Is the root %.5f close to %.5f?%n", root, PI/4); //
		
    }
}
